# -*- coding: utf-8 -*-
#
# This file is part of the EEHGMaster project
#
#
#
# Distributed under the terms of the GPL license.
# See LICENSE.txt for more info.

""" EEHG Master

Tango device server for EEHG operation for FERMI
"""

# PyTango imports
import tango
from tango import DebugIt
from tango.server import run
from tango.server import Device
from tango.server import attribute, command
from tango.server import device_property
from tango import AttrQuality, DispLevel, DevState
from tango import AttrWriteType, PipeWriteType
# Additional import
# PROTECTED REGION ID(EEHGMaster.additionnal_import) ENABLED START #
import threading
from time import sleep
import numpy as np
import utili as utl
# PROTECTED REGION END #    //  EEHGMaster.additionnal_import

__all__ = ["EEHGMaster", "main"]


class EEHGMaster(Device):
    """
    Tango device server for EEHG operation for FERMI

    **Properties:**

    - Device Property
        MasterFel1Dev
            - definisce il device da cui leggere la lunghezza d-onda del FEL
            - Type:'DevString'
        SclMagnetDev
            - Device da cui leggere l-energia del fascio di elettroni
            - Type:'DevString'
        DelayLineDev
            - Device da cui leggere il valore corrente R56\n (r561curr) della delay line
            - Type:'DevString'
        DispersiveMagDev
            - Device da cui leggere il valore corrente R56\n (r562curr) della sezione dispersiva
            - Type:'DevString'
        SeedMasterDev
            - Device che controlla il laser di seed
            - Type:'DevString'
        SeedCcd1Dev
            - Device da cui leggere HorSigma che diventa waist_s1
            - Type:'DevString'
        SeedCcd2Dev
            - Device da cui leggere HorSigma che diventa waist_s2
            - Type:'DevString'
        Modulator01Dev
            - Device per il controllo del modulatore n.1 di fel01
            - Type:'DevString'
        Modulator02Dev
            - Device of Modulator 2
            - Type:'DevString'
    """
    # PROTECTED REGION ID(EEHGMaster.class_variable) ENABLED START #
    # PROTECTED REGION END #    //  EEHGMaster.class_variable

    # -----------------
    # Device Properties
    # -----------------

    MasterFel1Dev = device_property(
        dtype='DevString',
        default_value="fel01/master/master_fel01.01"
    )

    SclMagnetDev = device_property(
        dtype='DevString',
        default_value="scl/magnet/b_scl.01"
    )

    DelayLineDev = device_property(
        dtype='DevString',
        default_value="fel01/magnet/delay_line_eehg_fel01.01"
    )

    DispersiveMagDev = device_property(
        dtype='DevString',
        default_value="fel01/magnet/dispersive_magnet_fel01.01"
    )

    SeedMasterDev = device_property(
        dtype='DevString',
        default_value="sl/master/seed_master_sl.01"
    )

    SeedCcd1Dev = device_property(
        dtype='DevString',
        default_value="sl/diagnostics/ccd_sl.07"
    )

    SeedCcd2Dev = device_property(
        dtype='DevString',
        default_value="sl/diagnostics/ccd_sl.28"
    )

    Modulator01Dev = device_property(
        dtype='DevString',
        default_value="fel01/id/id_mod_fel01.01"
    )

    Modulator02Dev = device_property(
        dtype='DevString',
        default_value="fel01/id/id_mod_fel01.02"
    )

    # ----------
    # Attributes
    # ----------

    fel_mode = attribute(
        dtype='DevShort',
        access=AttrWriteType.READ_WRITE,
        label="operation mode",
        max_value=3,
        min_value=0,
        memorized=True,
        hw_memorized=True,
        doc="FEL operation mode:\nEEHG\nHGHG1\nHGHG1*\nHGHG2\n\nSpecify how the FEL will be operated. \nValue:\n0 - EEHG is clear. \n1 - HGHG1, means that one use seed1 on mod1 to to standard HGHG. \n2 - HGHG1*, means that one use seed1 on mod2 to to standard HGHG. \n3 - HGHG2 means that seed2 is used on mod2 for standard HGHG. \n\nThis setting define the method to be used for calculating the required configuration",
    )

    n = attribute(
        dtype='DevShort',
        access=AttrWriteType.READ_WRITE,
        label="harmonic of the first seed",
        max_value=0,
        min_value=-1,
        memorized=True,
        doc="The harmonic of the first seed used in EEHG of in HGHG1 and HGHG1*",
    )

    m = attribute(
        dtype='DevShort',
        access=AttrWriteType.READ_WRITE,
        label="harmonic of the second seed",
        max_value=80,
        min_value=0,
        memorized=True,
        doc="The harmonic of the second seed used in EEHG or in HGHG2",
    )

    n_fixed = attribute(
        dtype='DevBoolean',
        access=AttrWriteType.READ_WRITE,
        label="n fixed",
        memorized=True,
        doc="keep n fix in the calculation",
    )

    m_fixed = attribute(
        dtype='DevBoolean',
        access=AttrWriteType.READ_WRITE,
        label="m fixed",
        memorized=True,
        doc="keep m fix in the calculation",
    )

    n_solution = attribute(
        dtype='DevShort',
        label="n solution",
        doc="Number of solutions found with the calculation",
    )

    i_solution = attribute(
        dtype='DevShort',
        access=AttrWriteType.READ_WRITE,
        label="Applied solution",
        doc="Applied solution from the ones founded",
    )

    prior_wl_seed = attribute(
        dtype='DevBoolean',
        access=AttrWriteType.READ_WRITE,
        label="prior_wl_seed",
        doc="priority to wl_seed closed to present",
    )

    prior_bunch = attribute(
        dtype='DevBoolean',
        access=AttrWriteType.READ_WRITE,
        label="prior_bunch",
        doc="Give the priority to the solution that has larger bunching",
    )

    bunc_fel = attribute(
        dtype='DevDouble',
        label="bunching fel",
        doc="Calculated value for the bunchign with the current settings at the specified wavelenght",
    )

    b_min = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
        label="bunching threshold",
        doc="This is the minimum value for the bunching that is used to create the list of solutions",
    )

    lambda_fel = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
        label="fel wavelength",
        max_value=25,
        min_value=5,
        memorized=True,
        doc="The FEL wavelength, it could be always in agreement with photon energy, if used",
    )

    lambda_fel_curr = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
        doc="The current values for thiese variable as read from the machine",
    )

    a1 = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
        label="A1 FEL parameter",
        memorized=True,
        doc="A1 FEL parameter",
    )

    emod1 = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
        label="emod1",
        unit="keV",
        doc="amplitude of the energy modulation from the first seed laser. It should be always connected to the seed laser energy and A1",
    )

    a1_curr = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
        label="a1_curr",
        doc="The current values for thiese variable as read from the machine",
    )

    emod1_curr = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
        label="emod1_curr",
        doc="The current values for thiese variable as read from the machine",
    )

    a1_fix = attribute(
        dtype='DevBoolean',
        access=AttrWriteType.READ_WRITE,
        label="a1_fix",
        memorized=True,
        doc="This is used to keep A1 and seed1 energy fixed in the optimization (default)",
    )

    a2 = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
        label="a2",
        doc="A2 EEHG parameter",
    )

    emod2 = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
        label="emod2",
        unit="keV",
        memorized=True,
        doc="amplitude of the energy modulation from the first seed laser. It should be always connected to the seed laser energy and A2",
    )

    a2_curr = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
        label="a2_curr",
        doc="The current values for thiese variable as read from the machine",
    )

    emod2_curr = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
        label="emod2_curr",
        doc="The current values for thiese variable as read from the machine",
    )

    a2_fix = attribute(
        dtype='DevBoolean',
        access=AttrWriteType.READ_WRITE,
        label="a2_fix",
        doc="This is used to keep A2 and seed2 energy fixed in the optimization, by default is false sto allow optimiziation of this parameter",
    )

    e_nrg_fel = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
        label="e_nrg_fel",
        unit="GeV",
        doc="the electron beam energy at the FEL used for all the calculations including the resonance for undulators",
    )

    e_dgamma_fel = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
        label="e_dgamma_fel",
        unit="keV",
        doc="the electron beam energy spread at the FEL used for all the bunching calculations",
    )

    e_compression = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
        label="e_compression",
        doc="the theoretical value of the compression of the beam. It is used for the calculation of the electron beam energy spread from the LH setting",
    )

    e_nrg_lh = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
        label="e_nrg_lh",
        unit="MeV",
        doc="the electron beam energy at the laser heater used for all the calculations of the LH indiced energy spread",
    )

    e_dgamma_lh = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
        label="e_dgamma_lh",
        unit="keV",
        doc="the electron beam energy spread at the laser heater",
    )

    r561 = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
        label="r561",
        unit="mm",
        max_value=10,
        min_value=0,
        doc="strength of the first dispersive section used in the calculations with respect to the machine R56 it is R56 =  (R56_machine-r56_off_abs)/(1+r56_off_perc/100)",
    )

    r561_curr = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
        label="r561_curr",
        doc="The current values for thiese variable as read from the machine",
    )

    r561_fix = attribute(
        dtype='DevBoolean',
        access=AttrWriteType.READ_WRITE,
        label="r561_fix",
        memorized=True,
        doc="boolean to know if this parameter can be changed during optimization",
    )

    r561_off_abs = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
        label="r561_off_abs",
        unit="um",
        max_value=10,
        min_value=0,
        memorized=True,
        doc="the value applied to the dispersion is changed according to r561_machine = r561* r561_off_perc + r561_off_abs",
    )

    r561_off_perc = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
        label="r561_off_perc",
        unit="%",
        max_value=1.6,
        min_value=0.4,
        memorized=True,
        doc="the value applied to the dispersion is changed according to r561_machine = r561*r561_off_perc + r561_off_abs",
    )

    r561_machine = attribute(
        dtype='DevDouble',
        label="r561_machine",
        doc="This is the value that is read from the machine and is calculated based on  r561_machine = r561*(1+r561_off_perc/100)+r561_off_abs",
    )

    r562 = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
        label="r562",
        unit="um",
        max_value=200,
        min_value=0,
        memorized=True,
        doc="strength of the first dispersive section",
    )

    r562_curr = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
        label="r562_curr",
        unit="um",
        doc="The current values for thiese variable as read from the machine",
    )

    r562_fix = attribute(
        dtype='DevBoolean',
        access=AttrWriteType.READ_WRITE,
        label="r562_fix",
        doc="boolean to know if this parameter can be changed during optimization.\nOnly 2 between (a1_fix,,a2_fix,R56_1_fix,R56_2_fix can and MUST be TRUE)",
    )

    r562_off_abs = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
        label="r562_off_abs",
        unit="um",
        max_value=300,
        min_value=0,
        memorized=True,
    )

    r562_off_perc = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
        label="r562_off_perc",
        unit="%",
        max_value=1.6,
        min_value=0.4,
        memorized=True,
        doc="the value applied to the dispersion is changed according to R56_machine = R56 * 56_off_perc + r56_off_abs",
    )

    r562_machine = attribute(
        dtype='DevDouble',
        label="r562_machine",
        doc="This is the value that is read from the machine and is calculated based on  R56_machine = R56*(1+r56_off_perc/100)+r56_off_abs",
    )

    lambda_s1 = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
        label="lambda_s1",
        unit="nm",
        max_value=261,
        min_value=259,
        doc="wavelength for the first seed",
    )

    dlambda_s1 = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
        label="dlambda_s1",
        unit="nm",
        memorized=True,
        doc="bandiwdth of the seed laser spectrum, it should be always be in agreement with the pulse length and the FL setting",
    )

    lambda_s1or2_fix = attribute(
        dtype='DevBoolean',
        access=AttrWriteType.READ_WRITE,
        label="lambda_s1or2_fix",
        memorized=True,
        doc="This has to be the negation of lambda_s2_fix ",
    )

    fl_s1 = attribute(
        dtype='DevBoolean',
        access=AttrWriteType.READ_WRITE,
        label="fl_s1",
        doc="specify if the seed is at the fourier limit. It need to be connected with the value for the pulse length. ",
    )

    dt_s1 = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
        label="dt_s1",
        unit="fs",
        doc="the length of the seed laser pulse. It should be alwais in agreement with the bandwidth and the fourier limit setting\nIf fl_s1 is true the numbers is obteined from the dlambda_s1 otherwise can be any number (longer than the FL)",
    )

    waist_s1 = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
        label="waist_s1",
        unit="um",
        memorized=True,
        doc="the size of the seed laser spot at the undulator",
    )

    waist_corr_s1 = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
        label="waist_corr_s1",
        memorized=True,
        doc="the correction to be used for the calculation of the seed laser field from the spot size, larger values means larger spot size (lower laser field). ",
    )

    nrg_s1 = attribute(
        dtype='DevDouble',
        label="nrg_s1",
        unit="uJ",
        max_value=50,
        min_value=0,
        doc="seed laser energy; This is linked to a1",
    )

    nrg_s1_curr = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
        label="nrg_s1_curr",
        unit="uJ",
        doc="The current values for thiese variable as read from the machine",
    )

    nrg_s1_fix = attribute(
        dtype='DevBoolean',
        access=AttrWriteType.READ_WRITE,
        label="nrg_s1_fix",
        memorized=True,
        doc="This is linked to a1_fix",
    )

    pow_s1 = attribute(
        dtype='DevDouble',
        label="pow_s1",
        unit="MW",
        doc="the peak power of the laser pulse it should be alwais in agreement with laser energy and pulse length:\nthis depends on nrg_s1 and dt_s1",
    )

    pow_s1_curr = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
        label="pow_s1_curr",
        doc="The current values for thiese variable as read from the machine",
    )

    lambda_s2 = attribute(
        dtype='DevDouble',
        label="lambda_s2",
        unit="nm",
        max_value=265,
        min_value=250,
        doc="wavelength for the second seed",
    )

    lambda_s2_curr = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
        label="lambda_s2_curr",
        doc="The current values for thiese variable as read from the machine",
    )

    dlambda_s2 = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
        label="dlambda_s2",
        unit="nm",
        memorized=True,
        doc="bandiwdth of the seed laser spectrum, it should be always be in agreement with the pulse length and the FL setting",
    )

    fl_s2 = attribute(
        dtype='DevBoolean',
        access=AttrWriteType.READ_WRITE,
        label="fl_s2",
        memorized=True,
        doc="specify if the seed is at the fourier limit. It need to be connected with the value for the pulse length. ",
    )

    dt_s2 = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
        label="dt_s2",
        unit="fs",
        memorized=True,
        doc="the length of the seed laser pulse. It should be alwais in agreement with the bandwidth and the fourier limit setting. When inserting a number it should not allow to put a number smaller than the FL\nIf fl_s2 is true the numbers is obteined from the dlambda_s2 otherwise can be any number (longer than the FL)",
    )

    waist_s2 = attribute(
        dtype='DevDouble',
        label="waist_s2",
        unit="um",
        doc="the size of the seed laser spot at the undulator",
    )

    waist_corr_s2 = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
        label="waist_corr_s2",
        memorized=True,
        doc="the correction to be used for the calculation of the seed laser field from the spot size, larger values means larger spot size (lower laser field). ",
    )

    nrg_s2 = attribute(
        dtype='DevDouble',
        label="nrg_s2",
        unit="uJ",
        max_value=50,
        min_value=0,
        doc="seed laser energy",
    )

    nrg_s2_curr = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
        label="nrg_s2_curr",
    )

    nrg_s2_fix = attribute(
        dtype='DevBoolean',
        access=AttrWriteType.READ_WRITE,
        label="nrg_s2_fix",
        memorized=True,
        doc="This is linked to a2_fix",
    )

    pow_s2 = attribute(
        dtype='DevDouble',
        label="pow_s2",
        unit="MW",
        doc="the peak power of the laser pulse it should be alwais in agreement with laser energy and pulse length\nthis depends on nrg_s2 and dt_s2",
    )

    pow_s2_curr = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
        label="pow_s2_curr",
        doc="current value",
    )

    lambda_s1_curr = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
        memorized=True,
    )

    picco_pref = attribute(
        dtype='DevShort',
        access=AttrWriteType.READ_WRITE,
        label="picco selection",
        max_value=2,
        min_value=0,
        memorized=True,
        doc="This is selection of R56 peak of EEHG: 0 left, 1 rigth, 2  both.",
    )

    bunc_sol = attribute(
        dtype=('DevLong64',),
        max_dim_x=32,
        label="bunching solution",
        doc="These are values of the bunching for all the solutions calculated by the script and ordererd according to the priorities and preferences",
    )

    # ---------------
    # General methods
    # ---------------

    def init_device(self):
        """Initialises the attributes and properties of the EEHGMaster."""
        Device.init_device(self)
        # PROTECTED REGION ID(EEHGMaster.init_device) ENABLED START #
        self._fel_mode = 0
        
        self.seed_dev_atts_toread = ['State', 'wavelength', 'bandwidth',
                                     'energy', 'energy_max', 'wavelength2',
                                     'bandwidth2', 'energy2', 'energy2_max',
                                     'wavelength_range', 'wavelength2_range']
        # definisco delle variabili di servizio per i limiti sup. dell'en. seed
        self.nrg_s1_range = 50.0
        self.nrg_s2_range = 20.0

        # valori startup di attributi
        self.n_value = -1
        self.m_value = 18
        self._n_fixed = True
        self._n_fixed = False
        self._n_solution = 0
        self._i_solution = 0
        self._lambda_fel = 12.0
        self._lambda_fel_curr = 13.3
        self._e_nrg_fel = 1.0
        self._e_dgamma_fel = float('Nan')
        self._e_compression = float('Nan')
        self._e_nrg_lh = 100.0
        self._e_dgamma_lh = float('Nan')
        self._picco_pref = 0
        
        self._r561 = float('Nan')
        self._r562 = float('Nan')
        self._r561_curr = 2
        self._r562_curr = 100
        self._r561_fix = True
        self._r561_off_abs = 0.0
        self._r561_off_perc = 1.0
        self._r561_machine = float('Nan')
        self._r562_fix = False
        self._r562_off_abs = 0.0
        self._r562_off_perc = 1.0
        self._r562_machine = float('Nan')
        
        self._lambda_s1 = 259.0
        self._lambda_s1_curr = 260.0
        self._dlambda_s1 = 1.0
        self._lambda_s1or2_fix = True
        self._fl_s1 = True
        self._waist_corr_s1 = float('Nan')
        self._nrg_s1_fix = True
        self._dt_s1 = 200e-15 #seed pulse
        self._nrg_s1 = float('Nan')
        self._nrg_s1_curr = 5.0
        self._pow_s1 = float('Nan')
        self._pow_s1_curr = float('Nan')
        
        self._lambda_s2 = 250.0
        self._lambda_s2_curr = 255.0
        self._dlambda_s2 = 1.0
        self._fl_s2 = True
        self._waist_corr_s2 = float('Nan')
        self._dt_s2 = 100e-15 #seed pulse
        self._nrg_s2 = float('Nan')
        self._nrg_s2_curr = 10.0
        self._nrg_s2_fix = False
        self._pow_s2 = float('Nan')
        self._pow_s2_curr = float('Nan')
        self._waist_s1 = 0.3
        self._waist_s2 = 0.2
        self._prior_wl_seed = True
        self._prior_bunch = False
        self._bunc_fel = float('Nan')
        self._b_min = float('Nan')
        
        self._a1 = float('Nan')
        self._a1_curr = float('Nan')
        self._a1_fix = True
        self._emod1 = float('Nan')
        self._emod1_curr = float('Nan')
        
        self._a2 = float('Nan')
        self._a2_curr = float('Nan')
        self._a2_fix = False
        self._emod2 = float('Nan')
        self._emod2_curr = float('Nan')

        self._bunc_sol = float('Nan')
                
        # inizializzo i devices:
        self.masterfel1_device = None
        if self.MasterFel1Dev:
            try:
                self.masterfel1_device = tango.DeviceProxy(self.MasterFel1Dev)
            except Exception as e:
                self.error_stream("Failed to connect to remote device: {}".format(e))
                self.set_state(tango.DevState.FAULT)
                self.set_status('failed to connect to remote device')
 
        self.mod01fel1_device = None
        if self.Modulator01Dev:
            try:
                self.mod01fel1_device = tango.DeviceProxy(self.Modulator01Dev)
            except Exception as e:
                self.error_stream("Failed to connect to remote device: {}".format(e))
                self.set_state(tango.DevState.FAULT)
                self.set_status('failed to connect to remote device')

        self.mod02fel1_device = None
        if self.Modulator02Dev:
            try:
                self.mod02fel1_device = tango.DeviceProxy(self.Modulator02Dev)
            except Exception as e:
                self.error_stream("Failed to connect to remote device: {}".format(e))
                self.set_state(tango.DevState.FAULT)
                self.set_status('failed to connect to remote device')

        self.sclmagnet_device = None
        if self.SclMagnetDev:
            try:
                self.sclmagnet_device = tango.DeviceProxy(self.SclMagnetDev)
            except Exception as e:
                self.error_stream("Failed to connect to remote device: {}".format(e))
                self.set_state(tango.DevState.FAULT)
                self.set_status('failed to connect to remote device')

        self.delayline_device = None
        if self.DelayLineDev:
            try:
                self.delayline_device = tango.DeviceProxy(self.DelayLineDev)
            except Exception as e:
                self.error_stream("Failed to connect to remote device: {}".format(e))
                self.set_state(tango.DevState.FAULT)
                self.set_status('failed to connect to remote device')

        self.dispmagnet_device = None 
        if self.DispersiveMagDev:
            try:
                self.dispmagnet_device = tango.DeviceProxy(self.DispersiveMagDev)
            except Exception as e:
                self.error_stream("Failed to connect to remote device: {}".format(e))
                self.set_state(tango.DevState.FAULT)
                self.set_status('failed to connect to remote device')

        self.seedmaster_device = None #prototipo
        if self.SeedMasterDev:
            try:
                self.seedmaster_device = tango.DeviceProxy(self.SeedMasterDev)
            except Exception as e:
                self.error_stream("Failed to connect to remote device: {}".format(e))
                self.set_state(tango.DevState.FAULT)
                self.set_status('failed to connect to remote device')

        self.seedccd1_device = None
        if self.SeedCcd1Dev:
            try:
                self.seedccd1_device = tango.DeviceProxy(self.SeedCcd1Dev)
            except Exception as e:
                self.error_stream("Failed to connect to remote device: {}".format(e))
                self.set_state(tango.DevState.FAULT)
                self.set_status('failed to connect to remote device')

        self.seedccd2_device = None
        if self.SeedCcd2Dev:
            try:
                self.seedccd2_device = tango.DeviceProxy(self.SeedCcd2Dev)
            except Exception as e:
                self.error_stream("Failed to connect to remote device: {}".format(e))
                self.set_state(tango.DevState.FAULT)
                self.set_status('failed to connect to remote device')

        self.set_state(tango.DevState.ON)
        self.set_status('Init ok, status ON')

        # PROTECTED REGION END #    //  EEHGMaster.init_device

    def always_executed_hook(self):
        """Method always executed before any TANGO command is executed."""
        # PROTECTED REGION ID(EEHGMaster.always_executed_hook) ENABLED START #
        # PROTECTED REGION END #    //  EEHGMaster.always_executed_hook

    def delete_device(self):
        """Hook to delete resources allocated in init_device.

        This method allows for any memory or other resources allocated in the
        init_device method to be released.  This method is called by the device
        destructor and by the device Init command.
        """
        # PROTECTED REGION ID(EEHGMaster.delete_device) ENABLED START #
        # PROTECTED REGION END #    //  EEHGMaster.delete_device
    # ------------------
    # Attributes methods
    # ------------------

    def read_fel_mode(self):
        # PROTECTED REGION ID(EEHGMaster.fel_mode_read) ENABLED START #
        """Return the fel_mode attribute."""
        return self._fel_mode
        # PROTECTED REGION END #    //  EEHGMaster.fel_mode_read

    def write_fel_mode(self, value):
        # PROTECTED REGION ID(EEHGMaster.fel_mode_write) ENABLED START #
        """Set the fel_mode attribute."""
        self._fel_mode = value
        # PROTECTED REGION END #    //  EEHGMaster.fel_mode_write

    def is_fel_mode_allowed(self, attr):
        # PROTECTED REGION ID(EEHGMaster.is_fel_mode_allowed) ENABLED START #
        if attr==attr.READ_REQ:
            return True
        else:
            return self.get_state() not in [DevState.RUNNING,DevState.MOVING,DevState.FAULT,DevState.ALARM,DevState.UNKNOWN]
        # PROTECTED REGION END #    //  EEHGMaster.is_fel_mode_allowed

    def read_n(self):
        # PROTECTED REGION ID(EEHGMaster.n_read) ENABLED START #
        """Return the n attribute."""
#        return self._n
        return self.n_value
    
        # PROTECTED REGION END #    //  EEHGMaster.n_read

    def write_n(self, value):
        # PROTECTED REGION ID(EEHGMaster.n_write) ENABLED START #
        """Set the n attribute."""
        self.n_value = value
        self.n.set_write_value(value)

        # PROTECTED REGION END #    //  EEHGMaster.n_write

    def is_n_allowed(self, attr):
        # PROTECTED REGION ID(EEHGMaster.is_n_allowed) ENABLED START #
        if attr==attr.READ_REQ:
            return self.get_state() not in [DevState.RUNNING,DevState.MOVING]
        else:
            return self.get_state() not in [DevState.RUNNING,DevState.MOVING]
        # PROTECTED REGION END #    //  EEHGMaster.is_n_allowed

    def read_m(self):
        # PROTECTED REGION ID(EEHGMaster.m_read) ENABLED START #
        """Return the m attribute."""
#        return self._m
        return self.m_value

        # PROTECTED REGION END #    //  EEHGMaster.m_read

    def write_m(self, value):
        # PROTECTED REGION ID(EEHGMaster.m_write) ENABLED START #
        """Set the m attribute."""
        self.m_value = value
        self.m.set_write_value(value)
        # PROTECTED REGION END #    //  EEHGMaster.m_write

    def is_m_allowed(self, attr):
        # PROTECTED REGION ID(EEHGMaster.is_m_allowed) ENABLED START #
        if attr==attr.READ_REQ:
            return True
        else:
            return self.get_state() not in [DevState.RUNNING,DevState.MOVING,DevState.FAULT,DevState.ALARM,DevState.UNKNOWN]
        # PROTECTED REGION END #    //  EEHGMaster.is_m_allowed

    def read_n_fixed(self):
        # PROTECTED REGION ID(EEHGMaster.n_fixed_read) ENABLED START #
        """Return the n_fixed attribute."""
        return self._n_fixed
        # PROTECTED REGION END #    //  EEHGMaster.n_fixed_read

    def write_n_fixed(self, value):
        # PROTECTED REGION ID(EEHGMaster.n_fixed_write) ENABLED START #
        """Set the n_fixed attribute."""
        pass
        # PROTECTED REGION END #    //  EEHGMaster.n_fixed_write

    def is_n_fixed_allowed(self, attr):
        # PROTECTED REGION ID(EEHGMaster.is_n_fixed_allowed) ENABLED START #
        if attr==attr.READ_REQ:
            return self.get_state() not in [DevState.RUNNING,DevState.MOVING]
        else:
            return self.get_state() not in [DevState.RUNNING,DevState.MOVING]
        # PROTECTED REGION END #    //  EEHGMaster.is_n_fixed_allowed

    def read_m_fixed(self):
        # PROTECTED REGION ID(EEHGMaster.m_fixed_read) ENABLED START #
        """Return the m_fixed attribute."""
        return self._m_fixed
        # PROTECTED REGION END #    //  EEHGMaster.m_fixed_read

    def write_m_fixed(self, value):
        # PROTECTED REGION ID(EEHGMaster.m_fixed_write) ENABLED START #
        """Set the m_fixed attribute."""
        pass
        # PROTECTED REGION END #    //  EEHGMaster.m_fixed_write

    def is_m_fixed_allowed(self, attr):
        # PROTECTED REGION ID(EEHGMaster.is_m_fixed_allowed) ENABLED START #
        if attr==attr.READ_REQ:
            return self.get_state() not in [DevState.RUNNING,DevState.MOVING]
        else:
            return self.get_state() not in [DevState.RUNNING,DevState.MOVING]
        # PROTECTED REGION END #    //  EEHGMaster.is_m_fixed_allowed

    def read_n_solution(self):
        # PROTECTED REGION ID(EEHGMaster.n_solution_read) ENABLED START #
        """Return the n_solution attribute."""
        return self._n_solution
        # PROTECTED REGION END #    //  EEHGMaster.n_solution_read

    def is_n_solution_allowed(self, attr):
        # PROTECTED REGION ID(EEHGMaster.is_n_solution_allowed) ENABLED START #
        return self.get_state() not in [DevState.RUNNING,DevState.MOVING]
        # PROTECTED REGION END #    //  EEHGMaster.is_n_solution_allowed

    def read_i_solution(self):
        # PROTECTED REGION ID(EEHGMaster.i_solution_read) ENABLED START #
        """Return the i_solution attribute."""
        return self._i_solution
        # PROTECTED REGION END #    //  EEHGMaster.i_solution_read

    def write_i_solution(self, value):
        # PROTECTED REGION ID(EEHGMaster.i_solution_write) ENABLED START #
        """Set the i_solution attribute."""
        self._i_solution = value
        # PROTECTED REGION END #    //  EEHGMaster.i_solution_write

    def is_i_solution_allowed(self, attr):
        # PROTECTED REGION ID(EEHGMaster.is_i_solution_allowed) ENABLED START #
        if attr==attr.READ_REQ:
            return self.get_state() not in [DevState.RUNNING,DevState.MOVING]
        else:
            return True
        # PROTECTED REGION END #    //  EEHGMaster.is_i_solution_allowed

    def read_prior_wl_seed(self):
        # PROTECTED REGION ID(EEHGMaster.prior_wl_seed_read) ENABLED START #
        """Return the prior_wl_seed attribute."""
        return self._prior_wl_seed
        # PROTECTED REGION END #    //  EEHGMaster.prior_wl_seed_read

    def write_prior_wl_seed(self, value):
        # PROTECTED REGION ID(EEHGMaster.prior_wl_seed_write) ENABLED START #
        """Set the prior_wl_seed attribute."""
        self._prior_wl_seed = value
        self._prior_bunch = not value
            
        # PROTECTED REGION END #    //  EEHGMaster.prior_wl_seed_write

    def is_prior_wl_seed_allowed(self, attr):
        # PROTECTED REGION ID(EEHGMaster.is_prior_wl_seed_allowed) ENABLED START #
        if attr==attr.READ_REQ:
            return True
        else:
            return self.get_state() not in [DevState.RUNNING,DevState.MOVING]
        # PROTECTED REGION END #    //  EEHGMaster.is_prior_wl_seed_allowed

    def read_prior_bunch(self):
        # PROTECTED REGION ID(EEHGMaster.prior_bunch_read) ENABLED START #
        """Return the prior_bunch attribute."""
        return self._prior_bunch
        # PROTECTED REGION END #    //  EEHGMaster.prior_bunch_read

    def write_prior_bunch(self, value):
        # PROTECTED REGION ID(EEHGMaster.prior_bunch_write) ENABLED START #
        """Set the prior_bunch attribute."""
        self._prior_bunch = value
        self._prior_wl_seed = not value
        # PROTECTED REGION END #    //  EEHGMaster.prior_bunch_write

    def is_prior_bunch_allowed(self, attr):
        # PROTECTED REGION ID(EEHGMaster.is_prior_bunch_allowed) ENABLED START #
        if attr==attr.READ_REQ:
            return True
        else:
            return self.get_state() not in [DevState.RUNNING,DevState.MOVING]
        # PROTECTED REGION END #    //  EEHGMaster.is_prior_bunch_allowed

    def read_bunc_fel(self):
        # PROTECTED REGION ID(EEHGMaster.bunc_fel_read) ENABLED START #
        """Return the bunc_fel attribute."""
        return self._bunc_fel
        # PROTECTED REGION END #    //  EEHGMaster.bunc_fel_read

    def read_b_min(self):
        # PROTECTED REGION ID(EEHGMaster.b_min_read) ENABLED START #
        """Return the b_min attribute."""
        return self._b_min
        # PROTECTED REGION END #    //  EEHGMaster.b_min_read

    def write_b_min(self, value):
        # PROTECTED REGION ID(EEHGMaster.b_min_write) ENABLED START #
        """Set the b_min attribute."""
        pass
        # PROTECTED REGION END #    //  EEHGMaster.b_min_write

    def read_lambda_fel(self):
        # PROTECTED REGION ID(EEHGMaster.lambda_fel_read) ENABLED START #
        """Return the lambda_fel attribute."""
        return self._lambda_fel
        # PROTECTED REGION END #    //  EEHGMaster.lambda_fel_read

    def write_lambda_fel(self, value):
        # PROTECTED REGION ID(EEHGMaster.lambda_fel_write) ENABLED START #
        """Set the lambda_fel attribute."""
        self._lambda_fel = value
        self.lambda_fel.set_write_value(value)
        # PROTECTED REGION END #    //  EEHGMaster.lambda_fel_write

    def is_lambda_fel_allowed(self, attr):
        # PROTECTED REGION ID(EEHGMaster.is_lambda_fel_allowed) ENABLED START #
        if attr==attr.READ_REQ:
            return True
        else:
            return self.get_state() not in [DevState.RUNNING,DevState.MOVING]
        # PROTECTED REGION END #    //  EEHGMaster.is_lambda_fel_allowed

    def read_lambda_fel_curr(self):
        # PROTECTED REGION ID(EEHGMaster.lambda_fel_curr_read) ENABLED START #
        """Return the lambda_fel_curr attribute."""
        return self._lambda_fel_curr
        # PROTECTED REGION END #    //  EEHGMaster.lambda_fel_curr_read

    def write_lambda_fel_curr(self, value):
        # PROTECTED REGION ID(EEHGMaster.lambda_fel_curr_write) ENABLED START #
        """Set the lambda_fel_curr attribute."""
        pass
        # PROTECTED REGION END #    //  EEHGMaster.lambda_fel_curr_write

    def is_lambda_fel_curr_allowed(self, attr):
        # PROTECTED REGION ID(EEHGMaster.is_lambda_fel_curr_allowed) ENABLED START #
        if attr==attr.READ_REQ:
            return True
        else:
            return self.get_state() not in [DevState.RUNNING,DevState.MOVING,DevState.FAULT]
        # PROTECTED REGION END #    //  EEHGMaster.is_lambda_fel_curr_allowed

    def read_a1(self):
        # PROTECTED REGION ID(EEHGMaster.a1_read) ENABLED START #
        """Return the a1 attribute."""
        return self._a1
        # PROTECTED REGION END #    //  EEHGMaster.a1_read

    def write_a1(self, value):
        # PROTECTED REGION ID(EEHGMaster.a1_write) ENABLED START #
        """Set the a1 attribute."""
        self._a1 = value
        self.a1.set_write_value(value)
        # PROTECTED REGION END #    //  EEHGMaster.a1_write

    def read_emod1(self):
        # PROTECTED REGION ID(EEHGMaster.emod1_read) ENABLED START #
        """Return the emod1 attribute."""
        return self._emod1
        # PROTECTED REGION END #    //  EEHGMaster.emod1_read

    def write_emod1(self, value):
        # PROTECTED REGION ID(EEHGMaster.emod1_write) ENABLED START #
        """Set the emod1 attribute."""
        pass
        # PROTECTED REGION END #    //  EEHGMaster.emod1_write

    def read_a1_curr(self):
        # PROTECTED REGION ID(EEHGMaster.a1_curr_read) ENABLED START #
        """Return the a1_curr attribute."""
        return self._a1_curr
        # PROTECTED REGION END #    //  EEHGMaster.a1_curr_read

    def write_a1_curr(self, value):
        # PROTECTED REGION ID(EEHGMaster.a1_curr_write) ENABLED START #
        """Set the a1_curr attribute."""
        pass
        # PROTECTED REGION END #    //  EEHGMaster.a1_curr_write

    def read_emod1_curr(self):
        # PROTECTED REGION ID(EEHGMaster.emod1_curr_read) ENABLED START #
        """Return the emod1_curr attribute."""
        return self._emod1_curr
        # PROTECTED REGION END #    //  EEHGMaster.emod1_curr_read

    def write_emod1_curr(self, value):
        # PROTECTED REGION ID(EEHGMaster.emod1_curr_write) ENABLED START #
        """Set the emod1_curr attribute."""
        pass
        # PROTECTED REGION END #    //  EEHGMaster.emod1_curr_write

    def read_a1_fix(self):
        # PROTECTED REGION ID(EEHGMaster.a1_fix_read) ENABLED START #
        """Return the a1_fix attribute."""
        return self._a1_fix
        # PROTECTED REGION END #    //  EEHGMaster.a1_fix_read

    def write_a1_fix(self, value):
        # PROTECTED REGION ID(EEHGMaster.a1_fix_write) ENABLED START #
        """Set the a1_fix attribute."""
        self._a1_fix = value
        self.a1_fix.set_write_value = value
        self._nrg_s1_fix = value 
        self._a2_fix = not value
        self._nrg_s2_fix = not value
        # PROTECTED REGION END #    //  EEHGMaster.a1_fix_write

    def is_a1_fix_allowed(self, attr):
        # PROTECTED REGION ID(EEHGMaster.is_a1_fix_allowed) ENABLED START #
        if attr==attr.READ_REQ:
            return True
        else:
            return self.get_state() not in [DevState.RUNNING,DevState.MOVING]
        # PROTECTED REGION END #    //  EEHGMaster.is_a1_fix_allowed

    def read_a2(self):
        # PROTECTED REGION ID(EEHGMaster.a2_read) ENABLED START #
        """Return the a2 attribute."""
        return self._a2
        # PROTECTED REGION END #    //  EEHGMaster.a2_read

    def write_a2(self, value):
        # PROTECTED REGION ID(EEHGMaster.a2_write) ENABLED START #
        """Set the a2 attribute."""
        self._a2 = value
        self.a2.set_write_value = value
        # PROTECTED REGION END #    //  EEHGMaster.a2_write

    def read_emod2(self):
        # PROTECTED REGION ID(EEHGMaster.emod2_read) ENABLED START #
        """Return the emod2 attribute."""
        return self._emod2
        # PROTECTED REGION END #    //  EEHGMaster.emod2_read

    def write_emod2(self, value):
        # PROTECTED REGION ID(EEHGMaster.emod2_write) ENABLED START #
        """Set the emod2 attribute."""
        self._emod2 = value
        # PROTECTED REGION END #    //  EEHGMaster.emod2_write

    def read_a2_curr(self):
        # PROTECTED REGION ID(EEHGMaster.a2_curr_read) ENABLED START #
        """Return the a2_curr attribute."""
        return self._a2_curr
        # PROTECTED REGION END #    //  EEHGMaster.a2_curr_read

    def write_a2_curr(self, value):
        # PROTECTED REGION ID(EEHGMaster.a2_curr_write) ENABLED START #
        """Set the a2_curr attribute."""
        pass
        # PROTECTED REGION END #    //  EEHGMaster.a2_curr_write

    def read_emod2_curr(self):
        # PROTECTED REGION ID(EEHGMaster.emod2_curr_read) ENABLED START #
        """Return the emod2_curr attribute."""
        return self._emod2_curr
        # PROTECTED REGION END #    //  EEHGMaster.emod2_curr_read

    def write_emod2_curr(self, value):
        # PROTECTED REGION ID(EEHGMaster.emod2_curr_write) ENABLED START #
        """Set the emod2_curr attribute."""
        pass
        # PROTECTED REGION END #    //  EEHGMaster.emod2_curr_write

    def read_a2_fix(self):
        # PROTECTED REGION ID(EEHGMaster.a2_fix_read) ENABLED START #
        """Return the a2_fix attribute."""
        return self._a2_fix
        # PROTECTED REGION END #    //  EEHGMaster.a2_fix_read

    def write_a2_fix(self, value):
        # PROTECTED REGION ID(EEHGMaster.a2_fix_write) ENABLED START #
        """Set the a2_fix attribute."""
        self._a2_fix = value
        self.a2_fix.set_write_value = value
        self._nrg_s2_fix = value
        self._a1_fix = not value
        self._nrg_s1_fix = not value

        # PROTECTED REGION END #    //  EEHGMaster.a2_fix_write

    def is_a2_fix_allowed(self, attr):
        # PROTECTED REGION ID(EEHGMaster.is_a2_fix_allowed) ENABLED START #
        if attr==attr.READ_REQ:
            return True
        else:
            return self.get_state() not in [DevState.RUNNING,DevState.MOVING]
        # PROTECTED REGION END #    //  EEHGMaster.is_a2_fix_allowed

    def read_e_nrg_fel(self):
        # PROTECTED REGION ID(EEHGMaster.e_nrg_fel_read) ENABLED START #
        """Return the e_nrg_fel attribute."""
        return self._e_nrg_fel
        # PROTECTED REGION END #    //  EEHGMaster.e_nrg_fel_read

    def write_e_nrg_fel(self, value):
        # PROTECTED REGION ID(EEHGMaster.e_nrg_fel_write) ENABLED START #
        """Set the e_nrg_fel attribute."""
        pass
        # PROTECTED REGION END #    //  EEHGMaster.e_nrg_fel_write

    def read_e_dgamma_fel(self):
        # PROTECTED REGION ID(EEHGMaster.e_dgamma_fel_read) ENABLED START #
        """Return the e_dgamma_fel attribute."""
        return self._e_dgamma_fel
        # PROTECTED REGION END #    //  EEHGMaster.e_dgamma_fel_read

    def write_e_dgamma_fel(self, value):
        # PROTECTED REGION ID(EEHGMaster.e_dgamma_fel_write) ENABLED START #
        """Set the e_dgamma_fel attribute."""
        pass
        # PROTECTED REGION END #    //  EEHGMaster.e_dgamma_fel_write

    def read_e_compression(self):
        # PROTECTED REGION ID(EEHGMaster.e_compression_read) ENABLED START #
        """Return the e_compression attribute."""
        return self._e_compression
        # PROTECTED REGION END #    //  EEHGMaster.e_compression_read

    def write_e_compression(self, value):
        # PROTECTED REGION ID(EEHGMaster.e_compression_write) ENABLED START #
        """Set the e_compression attribute."""
        pass
        # PROTECTED REGION END #    //  EEHGMaster.e_compression_write

    def read_e_nrg_lh(self):
        # PROTECTED REGION ID(EEHGMaster.e_nrg_lh_read) ENABLED START #
        """Return the e_nrg_lh attribute."""
        return self._e_nrg_lh
        # PROTECTED REGION END #    //  EEHGMaster.e_nrg_lh_read

    def write_e_nrg_lh(self, value):
        # PROTECTED REGION ID(EEHGMaster.e_nrg_lh_write) ENABLED START #
        """Set the e_nrg_lh attribute."""
        pass
        # PROTECTED REGION END #    //  EEHGMaster.e_nrg_lh_write

    def read_e_dgamma_lh(self):
        # PROTECTED REGION ID(EEHGMaster.e_dgamma_lh_read) ENABLED START #
        """Return the e_dgamma_lh attribute."""
        return self._e_dgamma_lh
        # PROTECTED REGION END #    //  EEHGMaster.e_dgamma_lh_read

    def write_e_dgamma_lh(self, value):
        # PROTECTED REGION ID(EEHGMaster.e_dgamma_lh_write) ENABLED START #
        """Set the e_dgamma_lh attribute."""
        pass
        # PROTECTED REGION END #    //  EEHGMaster.e_dgamma_lh_write

    def read_r561(self):
        # PROTECTED REGION ID(EEHGMaster.r561_read) ENABLED START #
        """Return the r561 attribute."""
        return self._r561
        # PROTECTED REGION END #    //  EEHGMaster.r561_read

    def write_r561(self, value):
        # PROTECTED REGION ID(EEHGMaster.r561_write) ENABLED START #
        """Set the r561 attribute."""
        self._r561 = value
        self.r561.set_write_value(value)
        # PROTECTED REGION END #    //  EEHGMaster.r561_write

    def read_r561_curr(self):
        # PROTECTED REGION ID(EEHGMaster.r561_curr_read) ENABLED START #
        """Return the r561_curr attribute."""
        return self._r561_curr
        # PROTECTED REGION END #    //  EEHGMaster.r561_curr_read

    def write_r561_curr(self, value):
        # PROTECTED REGION ID(EEHGMaster.r561_curr_write) ENABLED START #
        """Set the r561_curr attribute."""
        pass
        # PROTECTED REGION END #    //  EEHGMaster.r561_curr_write

    def read_r561_fix(self):
        # PROTECTED REGION ID(EEHGMaster.r561_fix_read) ENABLED START #
        """Return the r561_fix attribute."""
        return self._r561_fix
        # PROTECTED REGION END #    //  EEHGMaster.r561_fix_read

    def write_r561_fix(self, value):
        # PROTECTED REGION ID(EEHGMaster.r561_fix_write) ENABLED START #
        """Set the r561_fix attribute."""
        self._r561_fix = value
        self._r562_fix = not value
        # PROTECTED REGION END #    //  EEHGMaster.r561_fix_write

    def read_r561_off_abs(self):
        # PROTECTED REGION ID(EEHGMaster.r561_off_abs_read) ENABLED START #
        """Return the r561_off_abs attribute."""
        return self._r561_off_abs
        # PROTECTED REGION END #    //  EEHGMaster.r561_off_abs_read

    def write_r561_off_abs(self, value):
        # PROTECTED REGION ID(EEHGMaster.r561_off_abs_write) ENABLED START #
        """Set the r561_off_abs attribute."""
        self._r561_off_abs = value
        # PROTECTED REGION END #    //  EEHGMaster.r561_off_abs_write

    def read_r561_off_perc(self):
        # PROTECTED REGION ID(EEHGMaster.r561_off_perc_read) ENABLED START #
        """Return the r561_off_perc attribute."""
        return self._r561_off_perc
        # PROTECTED REGION END #    //  EEHGMaster.r561_off_perc_read

    def write_r561_off_perc(self, value):
        # PROTECTED REGION ID(EEHGMaster.r561_off_perc_write) ENABLED START #
        """Set the r561_off_perc attribute."""
        pass
        # PROTECTED REGION END #    //  EEHGMaster.r561_off_perc_write

    def read_r561_machine(self):
        # PROTECTED REGION ID(EEHGMaster.r561_machine_read) ENABLED START #
        """Return the r561_machine attribute."""
        return self._r561_machine
        # PROTECTED REGION END #    //  EEHGMaster.r561_machine_read

    def read_r562(self):
        # PROTECTED REGION ID(EEHGMaster.r562_read) ENABLED START #
        """Return the r562 attribute."""
        return self._r562
        # PROTECTED REGION END #    //  EEHGMaster.r562_read

    def write_r562(self, value):
        # PROTECTED REGION ID(EEHGMaster.r562_write) ENABLED START #
        """Set the r562 attribute."""
        self._r562 = value
        self.r562.set_write_value(value)
    
        # PROTECTED REGION END #    //  EEHGMaster.r562_write

    def read_r562_curr(self):
        # PROTECTED REGION ID(EEHGMaster.r562_curr_read) ENABLED START #
        """Return the r562_curr attribute."""
        return self._r562_curr
        # PROTECTED REGION END #    //  EEHGMaster.r562_curr_read

    def write_r562_curr(self, value):
        # PROTECTED REGION ID(EEHGMaster.r562_curr_write) ENABLED START #
        """Set the r562_curr attribute."""
        pass
        # PROTECTED REGION END #    //  EEHGMaster.r562_curr_write

    def read_r562_fix(self):
        # PROTECTED REGION ID(EEHGMaster.r562_fix_read) ENABLED START #
        """Return the r562_fix attribute."""
        return self._r562_fix
        # PROTECTED REGION END #    //  EEHGMaster.r562_fix_read

    def write_r562_fix(self, value):
        # PROTECTED REGION ID(EEHGMaster.r562_fix_write) ENABLED START #
        """Set the r562_fix attribute."""
        self._r562_fix = value
        self._r561_fix = not value
        # PROTECTED REGION END #    //  EEHGMaster.r562_fix_write

    def read_r562_off_abs(self):
        # PROTECTED REGION ID(EEHGMaster.r562_off_abs_read) ENABLED START #
        """Return the r562_off_abs attribute."""
        return self._r562_off_abs
        # PROTECTED REGION END #    //  EEHGMaster.r562_off_abs_read

    def write_r562_off_abs(self, value):
        # PROTECTED REGION ID(EEHGMaster.r562_off_abs_write) ENABLED START #
        """Set the r562_off_abs attribute."""
        pass
        # PROTECTED REGION END #    //  EEHGMaster.r562_off_abs_write

    def read_r562_off_perc(self):
        # PROTECTED REGION ID(EEHGMaster.r562_off_perc_read) ENABLED START #
        """Return the r562_off_perc attribute."""
        return self._r562_off_perc
        # PROTECTED REGION END #    //  EEHGMaster.r562_off_perc_read

    def write_r562_off_perc(self, value):
        # PROTECTED REGION ID(EEHGMaster.r562_off_perc_write) ENABLED START #
        """Set the r562_off_perc attribute."""
        pass
        # PROTECTED REGION END #    //  EEHGMaster.r562_off_perc_write

    def read_r562_machine(self):
        # PROTECTED REGION ID(EEHGMaster.r562_machine_read) ENABLED START #
        """Return the r562_machine attribute."""
        return self._r562_machine
        # PROTECTED REGION END #    //  EEHGMaster.r562_machine_read

    def read_lambda_s1(self):
        # PROTECTED REGION ID(EEHGMaster.lambda_s1_read) ENABLED START #
        """Return the lambda_s1 attribute."""
        return self._lambda_s1
        # PROTECTED REGION END #    //  EEHGMaster.lambda_s1_read

    def write_lambda_s1(self, value):
        # PROTECTED REGION ID(EEHGMaster.lambda_s1_write) ENABLED START #
        """Set the lambda_s1 attribute."""
        pass
        self._lambda_s1 = value
        self.lambda_s1.set_write_value(value)
        # PROTECTED REGION END #    //  EEHGMaster.lambda_s1_write

    def read_dlambda_s1(self):
        # PROTECTED REGION ID(EEHGMaster.dlambda_s1_read) ENABLED START #
        """Return the dlambda_s1 attribute."""
        return self._dlambda_s1
        # PROTECTED REGION END #    //  EEHGMaster.dlambda_s1_read

    def write_dlambda_s1(self, value):
        # PROTECTED REGION ID(EEHGMaster.dlambda_s1_write) ENABLED START #
        """Set the dlambda_s1 attribute."""
        self._dlambda_s1 = value
        # PROTECTED REGION END #    //  EEHGMaster.dlambda_s1_write

    def read_lambda_s1or2_fix(self):
        # PROTECTED REGION ID(EEHGMaster.lambda_s1or2_fix_read) ENABLED START #
        """Return the lambda_s1or2_fix attribute."""
        return self._lambda_s1or2_fix
        # PROTECTED REGION END #    //  EEHGMaster.lambda_s1or2_fix_read

    def write_lambda_s1or2_fix(self, value):
        # PROTECTED REGION ID(EEHGMaster.lambda_s1or2_fix_write) ENABLED START #
        """Set the lambda_s1or2_fix attribute."""
        self._lambda_s1or2_fix = value
        # PROTECTED REGION END #    //  EEHGMaster.lambda_s1or2_fix_write

    def read_fl_s1(self):
        # PROTECTED REGION ID(EEHGMaster.fl_s1_read) ENABLED START #
        """Return the fl_s1 attribute."""
        return self._fl_s1
        # PROTECTED REGION END #    //  EEHGMaster.fl_s1_read

    def write_fl_s1(self, value):
        # PROTECTED REGION ID(EEHGMaster.fl_s1_write) ENABLED START #
        """Set the fl_s1 attribute."""
        pass
        # PROTECTED REGION END #    //  EEHGMaster.fl_s1_write

    def read_dt_s1(self):
        # PROTECTED REGION ID(EEHGMaster.dt_s1_read) ENABLED START #
        """Return the dt_s1 attribute."""
        return self._dt_s1
        # PROTECTED REGION END #    //  EEHGMaster.dt_s1_read

    def write_dt_s1(self, value):
        # PROTECTED REGION ID(EEHGMaster.dt_s1_write) ENABLED START #
        """Set the dt_s1 attribute."""
        self._dt_s1 = value
        pass
        # PROTECTED REGION END #    //  EEHGMaster.dt_s1_write

    def read_waist_s1(self):
        # PROTECTED REGION ID(EEHGMaster.waist_s1_read) ENABLED START #
        """Return the waist_s1 attribute."""
        return self._waist_s1
        # PROTECTED REGION END #    //  EEHGMaster.waist_s1_read

    def write_waist_s1(self, value):
        # PROTECTED REGION ID(EEHGMaster.waist_s1_write) ENABLED START #
        """Set the waist_s1 attribute."""
        self._waist_s1 = value
        self.waist_s1.set_write_value(value)
        # PROTECTED REGION END #    //  EEHGMaster.waist_s1_write

    def read_waist_corr_s1(self):
        # PROTECTED REGION ID(EEHGMaster.waist_corr_s1_read) ENABLED START #
        """Return the waist_corr_s1 attribute."""
        return self._waist_corr_s1
        # PROTECTED REGION END #    //  EEHGMaster.waist_corr_s1_read

    def write_waist_corr_s1(self, value):
        # PROTECTED REGION ID(EEHGMaster.waist_corr_s1_write) ENABLED START #
        """Set the waist_corr_s1 attribute."""
        pass
        # PROTECTED REGION END #    //  EEHGMaster.waist_corr_s1_write

    def read_nrg_s1(self):
        # PROTECTED REGION ID(EEHGMaster.nrg_s1_read) ENABLED START #
        """Return the nrg_s1 attribute."""
        return self._nrg_s1
        # PROTECTED REGION END #    //  EEHGMaster.nrg_s1_read

    def read_nrg_s1_curr(self):
        # PROTECTED REGION ID(EEHGMaster.nrg_s1_curr_read) ENABLED START #
        """Return the nrg_s1_curr attribute."""
        return self._nrg_s1_curr
        # PROTECTED REGION END #    //  EEHGMaster.nrg_s1_curr_read

    def write_nrg_s1_curr(self, value):
        # PROTECTED REGION ID(EEHGMaster.nrg_s1_curr_write) ENABLED START #
        """Set the nrg_s1_curr attribute."""
        self.nrg_s1_curr.set_write_value(value)
        self._nrg_s1_curr = value
        pass
        # PROTECTED REGION END #    //  EEHGMaster.nrg_s1_curr_write

    def read_nrg_s1_fix(self):
        # PROTECTED REGION ID(EEHGMaster.nrg_s1_fix_read) ENABLED START #
        """Return the nrg_s1_fix attribute."""
        return self._nrg_s1_fix
        # PROTECTED REGION END #    //  EEHGMaster.nrg_s1_fix_read

    def write_nrg_s1_fix(self, value):
        # PROTECTED REGION ID(EEHGMaster.nrg_s1_fix_write) ENABLED START #
        """Set the nrg_s1_fix attribute."""
        self._nrg_s1_fix = value
#        self.write_a1_fix(value)
        self._a1_fix = value
        self._nrg_s1_fix = value 
        self._a2_fix = not value
        self._nrg_s2_fix = not value
        # PROTECTED REGION END #    //  EEHGMaster.nrg_s1_fix_write

    def read_pow_s1(self):
        # PROTECTED REGION ID(EEHGMaster.pow_s1_read) ENABLED START #
        """Return the pow_s1 attribute."""
        return self._pow_s1
        # PROTECTED REGION END #    //  EEHGMaster.pow_s1_read

    def read_pow_s1_curr(self):
        # PROTECTED REGION ID(EEHGMaster.pow_s1_curr_read) ENABLED START #
        """Return the pow_s1_curr attribute."""
        return self._pow_s1_curr
        # PROTECTED REGION END #    //  EEHGMaster.pow_s1_curr_read

    def write_pow_s1_curr(self, value):
        # PROTECTED REGION ID(EEHGMaster.pow_s1_curr_write) ENABLED START #
        """Set the pow_s1_curr attribute."""
        pass
        # PROTECTED REGION END #    //  EEHGMaster.pow_s1_curr_write

    def read_lambda_s2(self):
        # PROTECTED REGION ID(EEHGMaster.lambda_s2_read) ENABLED START #
        """Return the lambda_s2 attribute."""
        return self._lambda_s2
        # PROTECTED REGION END #    //  EEHGMaster.lambda_s2_read

    def read_lambda_s2_curr(self):
        # PROTECTED REGION ID(EEHGMaster.lambda_s2_curr_read) ENABLED START #
        """Return the lambda_s2_curr attribute."""
        return self._lambda_s2_curr
        # PROTECTED REGION END #    //  EEHGMaster.lambda_s2_curr_read

    def write_lambda_s2_curr(self, value):
        # PROTECTED REGION ID(EEHGMaster.lambda_s2_curr_write) ENABLED START #
        """Set the lambda_s2_curr attribute."""
        self._lambda_s2_curr = value
        # PROTECTED REGION END #    //  EEHGMaster.lambda_s2_curr_write

    def read_dlambda_s2(self):
        # PROTECTED REGION ID(EEHGMaster.dlambda_s2_read) ENABLED START #
        """Return the dlambda_s2 attribute."""
        return self._dlambda_s2
        # PROTECTED REGION END #    //  EEHGMaster.dlambda_s2_read

    def write_dlambda_s2(self, value):
        # PROTECTED REGION ID(EEHGMaster.dlambda_s2_write) ENABLED START #
        """Set the dlambda_s2 attribute."""
        self._dlambda_s2 = value
        # PROTECTED REGION END #    //  EEHGMaster.dlambda_s2_write

    def read_fl_s2(self):
        # PROTECTED REGION ID(EEHGMaster.fl_s2_read) ENABLED START #
        """Return the fl_s2 attribute."""
        return self._fl_s2
        # PROTECTED REGION END #    //  EEHGMaster.fl_s2_read

    def write_fl_s2(self, value):
        # PROTECTED REGION ID(EEHGMaster.fl_s2_write) ENABLED START #
        """Set the fl_s2 attribute."""
        pass
        # PROTECTED REGION END #    //  EEHGMaster.fl_s2_write

    def read_dt_s2(self):
        # PROTECTED REGION ID(EEHGMaster.dt_s2_read) ENABLED START #
        """Return the dt_s2 attribute."""
        return self._dt_s2
        # PROTECTED REGION END #    //  EEHGMaster.dt_s2_read

    def write_dt_s2(self, value):
        # PROTECTED REGION ID(EEHGMaster.dt_s2_write) ENABLED START #
        """Set the dt_s2 attribute."""
        self._dt_s2 = value #seed pulse

        # PROTECTED REGION END #    //  EEHGMaster.dt_s2_write

    def read_waist_s2(self):
        # PROTECTED REGION ID(EEHGMaster.waist_s2_read) ENABLED START #
        """Return the waist_s2 attribute."""
        return self._waist_s2
        # PROTECTED REGION END #    //  EEHGMaster.waist_s2_read

    def read_waist_corr_s2(self):
        # PROTECTED REGION ID(EEHGMaster.waist_corr_s2_read) ENABLED START #
        """Return the waist_corr_s2 attribute."""
        return self._waist_corr_s2
        # PROTECTED REGION END #    //  EEHGMaster.waist_corr_s2_read

    def write_waist_corr_s2(self, value):
        # PROTECTED REGION ID(EEHGMaster.waist_corr_s2_write) ENABLED START #
        """Set the waist_corr_s2 attribute."""
        self._waist_corr_s2 = value
        # PROTECTED REGION END #    //  EEHGMaster.waist_corr_s2_write

    def read_nrg_s2(self):
        # PROTECTED REGION ID(EEHGMaster.nrg_s2_read) ENABLED START #
        """Return the nrg_s2 attribute."""
        return self._nrg_s2
        # PROTECTED REGION END #    //  EEHGMaster.nrg_s2_read

    def read_nrg_s2_curr(self):
        # PROTECTED REGION ID(EEHGMaster.nrg_s2_curr_read) ENABLED START #
        """Return the nrg_s2_curr attribute."""
        return self._nrg_s2_curr
        # PROTECTED REGION END #    //  EEHGMaster.nrg_s2_curr_read

    def write_nrg_s2_curr(self, value):
        # PROTECTED REGION ID(EEHGMaster.nrg_s2_curr_write) ENABLED START #
        """Set the nrg_s2_curr attribute."""
        pass
        # PROTECTED REGION END #    //  EEHGMaster.nrg_s2_curr_write

    def read_nrg_s2_fix(self):
        # PROTECTED REGION ID(EEHGMaster.nrg_s2_fix_read) ENABLED START #
        """Return the nrg_s2_fix attribute."""
        return self._nrg_s2_fix
        # PROTECTED REGION END #    //  EEHGMaster.nrg_s2_fix_read

    def write_nrg_s2_fix(self, value):
        # PROTECTED REGION ID(EEHGMaster.nrg_s2_fix_write) ENABLED START #
        """Set the nrg_s2_fix attribute."""
        self._a1_fix = not value
        self._nrg_s1_fix = not value 
        self._a2_fix = value
        self._nrg_s2_fix = value
        # PROTECTED REGION END #    //  EEHGMaster.nrg_s2_fix_write

    def read_pow_s2(self):
        # PROTECTED REGION ID(EEHGMaster.pow_s2_read) ENABLED START #
        """Return the pow_s2 attribute."""
        return self._pow_s2
        # PROTECTED REGION END #    //  EEHGMaster.pow_s2_read

    def read_pow_s2_curr(self):
        # PROTECTED REGION ID(EEHGMaster.pow_s2_curr_read) ENABLED START #
        """Return the pow_s2_curr attribute."""
        return self._pow_s2_curr
        # PROTECTED REGION END #    //  EEHGMaster.pow_s2_curr_read

    def write_pow_s2_curr(self, value):
        # PROTECTED REGION ID(EEHGMaster.pow_s2_curr_write) ENABLED START #
        """Set the pow_s2_curr attribute."""
        pass
        # PROTECTED REGION END #    //  EEHGMaster.pow_s2_curr_write

    def read_lambda_s1_curr(self):
        # PROTECTED REGION ID(EEHGMaster.lambda_s1_curr_read) ENABLED START #
        """Return the lambda_s1_curr attribute."""
        return self._lambda_s1_curr
        # PROTECTED REGION END #    //  EEHGMaster.lambda_s1_curr_read

    def write_lambda_s1_curr(self, value):
        # PROTECTED REGION ID(EEHGMaster.lambda_s1_curr_write) ENABLED START #
        """Set the lambda_s1_curr attribute."""
        self._lambda_s1_curr = value
        # PROTECTED REGION END #    //  EEHGMaster.lambda_s1_curr_write

    def read_picco_pref(self):
        # PROTECTED REGION ID(EEHGMaster.picco_pref_read) ENABLED START #
        """Return the picco_pref attribute."""
        return self._picco_pref
        # PROTECTED REGION END #    //  EEHGMaster.picco_pref_read

    def write_picco_pref(self, value):
        # PROTECTED REGION ID(EEHGMaster.picco_pref_write) ENABLED START #
        """Set the picco_pref attribute."""
        self._picco_pref = value
        pass
        # PROTECTED REGION END #    //  EEHGMaster.picco_pref_write

    def read_bunc_sol(self):
        # PROTECTED REGION ID(EEHGMaster.bunc_sol_read) ENABLED START #
        """Return the bunc_sol attribute."""
        return self._bunc_sol
        # PROTECTED REGION END #    //  EEHGMaster.bunc_sol_read

    # --------
    # Commands
    # --------

    @command(
    )
    @DebugIt()
    def Calculate(self):
        # PROTECTED REGION ID(EEHGMaster.Calculate) ENABLED START #
        """
        :return:None
        """
        thread = threading.Thread(target=self.run_thread, daemon=True)
        thread.start()

    def run_thread(self):
        # inserire qui i caolcoli di utili
        self.set_state(tango.DevState.RUNNING)
        self.set_status('Started calculation')

        #n_curr = -1 #andrebbe letto dal server EEHG come valore di n
        #m_curr = 21 #andrebbe letto dal server EEHG come valore di m
        #n_rng = [-1,0] # puo' essere anche [-n, 0], n = 1, 2, 3 ... ma all' aumentare di n il tempo di calcolo aumenta
        #m_rng = [1,50]
        n_rng = [int(self.n.get_properties().min_value), int(self.n.get_properties().max_value)] # puo' essere anche [-n, 0], n = 1, 2, 3 ... ma all' aumentare di n il tempo di calcolo aumenta
        m_rng = [int(self.m.get_properties().min_value), int(self.m.get_properties().max_value)]

        #picco = 'left' # can be also 'right' or ' both'
        #picco = 'left' # can be also 'right' or ' both'
        #i_solution = 0 # indice della soluzione da applicare. per difetto e' la prima trovata
        picchi = ['left', 'right', ' both']
        picco = picchi[self._picco_pref]
        """
        calcola configurazioni usando utili
        """
        printout = False

        ebeam = utl.ElectronBeam(energy = self._e_nrg_fel * 1e9, energyspread = 100e3,
                                   current = 500.0, emittance_n = 1e-6, beamsize=100e-6)

        mod1 = utl.Undulator(lw = 0.10, lung = 3.0, helical = False, aw = 1.0)
        mod2 = utl.Undulator(lw = 0.113, lung = 1.469, helical = False, aw = 1.0)

        seed1 = utl.Laser(wl = self._lambda_s1_curr * 1e-9, bdw = self._dlambda_s1*1.0e-9,
                           nrg = self._nrg_s1_curr * 1e-6, waist = self._waist_s1 * 1e-3,
                            wl_range = [self._lambda_s1_curr * 1e-9, self._lambda_s1_curr * 1e-9])
#                            wl_range matenuto fisso attorno al seed attuale
#                            wl_range = [self.lambda_s1.get_properties().min_value * 1e-9,
#                                        self.lambda_s1.get_properties().max_value * 1e-9])

        seed1.set_duration(self._dt_s1)
        seed1.nrg_range = [0.0, self.nrg_s1_range * 1e-6]
        
        mod1.calc_aw(seed1.wl, ebeam.energy)

        risu = seed1.calc_a_from_seedpower(ebeam, mod1, arange = True, printout = printout)
        A1_range = risu[0]
        A1_curr = seed1.calc_a_from_seedpower(ebeam, mod1, arange = False, printout = printout)[0][0]

        seed2 = utl.Laser(wl = self._lambda_s2_curr * 1e-9, bdw = self._dlambda_s2 * 1.0e-9,
                            nrg = self._nrg_s2_curr * 1e-6, waist = self._waist_s2 * 1e-3,
                            wl_range = [float(self.lambda_s2.get_properties().min_value) * 1e-9,
                                        float(self.lambda_s2.get_properties().max_value) * 1e-9])
#                            wl_range = [240 * 1e-9, 265 * 1e-9])
#                            wl_range = [lambda_s2_range[0]*1e-9, lambda_s2_range[1]*1e-9]) self.lambda_s2.get_properties().max_value

        seed2.set_duration(self._dt_s2)
        seed2.nrg_range = [0.0, self.nrg_s2_range] # andrebbe aggiustato sulla di quanto scritto nel seed master 

        mod2.calc_aw(seed2.wl, ebeam.energy)
        
        risu = seed2.calc_a_from_seedpower(ebeam, mod2, arange = True, printout = printout)
        A2_range = risu[0]
        A2_curr = seed2.calc_a_from_seedpower(ebeam, mod2, arange = False, printout = printout)[0][0]

        r561 = utl.DispersiveSection(r56=0.0, lung = 3.0, dx=0.0, dt=0.0, r56_range=[0.0, 10e-3])
        R561_range = r561.r56_range
        
        r562 = utl.DispersiveSection(r56=0.0, lung = 1.2, dx=0.0, dt=0.0, 
                                     r56_range=[0.0, float(self.r562.get_properties().max_value) *1e-6]) #dovremmo aggiornare il valore massimo di r56 con la dipendenza dall'energia
        R562_range = r562.r56_range

#        eehg_ora = utili.EEHGsetup(A1 = A1_curr, A2 = A2_curr, n= n_curr, m = m_curr, R56_1 = r561curr*1e-3, R56_2 = r562curr*1e-6)
        try:
            eehg_ora = utl.EEHGsetup(A1 = A1_curr, A2 = A2_curr, 
                                   n = self.n_value, m = self.m_value,
                                   R56_1 = self._r561_curr * 1e-3,
                                   R56_2 = self._r562_curr * 1e-6)
        except:
            self.set_state(tango.DevState.FAULT)
            self.set_status('Calculation failed.')
            return

        par_err = np.array([5.0,5.0,2.0,2.0])

        try:
            [n_sol,m_sol,wls1_sol,wls2_sol] = eehg_ora.calc_wls2_n_m(self._lambda_fel * 1e-9,
                    seed1, seed2, m_range = m_rng,
                    n_range = n_rng, n_goal = -1, min_wls2_chg = True,
                    printout = printout)
# 
            solu = eehg_ora.calc_eehg_conf(ebeam, n_sol, m_sol, wls1_sol, wls2_sol,
                fix1 = 'A1', fix2 = 'R561', peak = picco, par_err = par_err, 
                printout = printout)
        except:
            self.set_state(tango.DevState.FAULT)
            self.set_status('Calculation failed.')
            return
        
        # variabili locali
        A1_sol = solu[0]
        A2_sol = solu[1]
        R561_sol = solu[2] * self._r561_off_perc + self._r561_off_abs * 1e-3
        R562_sol = solu[3] * self._r562_off_perc + self._r562_off_abs * 1e-6
        Wls1_sol = solu[4]
        Wls2_sol = solu[5]
        N_sol = solu[6]
        M_sol = solu[7]
        Bmax_sol = solu[8]
        B_var = solu[9]

        indTutto = np.where(Bmax_sol>0.00001)
        
        indA1 = np.where( (A1_sol>=A1_range[0]) & (A1_sol<=A1_range[1]))
        indA2 = np.where( (A2_sol>=A2_range[0]) & (A2_sol<=A2_range[1]))
        indR561 = np.where( (R561_sol>=R561_range[0]) & (R561_sol<=R561_range[1]))
        indR562 = np.where( (R562_sol>=R562_range[0]) & (R562_sol<=R562_range[1]))
        
        indOK = np.intersect1d(indTutto,indA1)
        indOK = np.intersect1d(indOK,indA2)
        indOK = np.intersect1d(indOK,indR561)
        indOK = np.intersect1d(indOK,indR562)

        if len(indOK)>0:
            #print('\033[92m'+'ho trovato almeno una soluzione valida\n'+'\033[0m')
            pass
        else:
#            print('\033[93m'+'non ho trovato nessuna soluzione valida\n'+'\033[0m')
#            solucheck = False
            self.set_state(tango.DevState.FAULT)
            self.set_status('Calculation do not found solutions.')
            return


        '''
        soluzione da applicare i_solution
        '''
        A1_sol_ap = solu[0][self._i_solution] # non si applica! seed1 fixed energy
        A2_sol_ap = solu[1][self._i_solution] # da convertire in energia
# energia calcolata per seed2 in uJ da applicare al device sl/master/seed_master_sl.01/energy2
        self._nrg_s2 = seed2.calc_seedpower_from_a(A2_sol_ap, ebeam, mod2, printout = printout)[1]*1e6 

        self._r561 = (solu[2][self._i_solution] * self._r561_off_perc * 1e3
                      + self._r561_off_abs) # non si applica perche' r561 e' fissa
#        R562_sol_ap = solu[3][self._i_solution] # da applicare direttamente al device fel01/magnet/dispersive_magnet_fel01.01/R56
        self._r562 = (solu[3][self._i_solution]  * self._r562_off_perc * 1e6
                      + self._r562_off_abs)
        self._lambda_s1 = solu[4][self._i_solution] * 1e9 # non si applica! seed1 fixed wavelegth 
#        Wls2_sol_ap = solu[5][self._i_solution] # da applicare direttamente al device sl/master/seed_master_sl.01/wavelength2
        self._lambda_s2 = solu[5][self._i_solution]  * 1e9
#        N_so_apl = solu[6][self._i_solution] # da scivere nel server sul valore di n
        self.write_n(solu[6][self._i_solution])
#        M_sol_ap = solu[7][self._i_solution] # da scivere nel server sul valore di m
        self.write_m(solu[7][self._i_solution])
#        Bmax_sol_ap = solu[8][self._i_solution] # da scivere nel server su bunch_fel
#        bunching_array = solu[8]
#        print(bunching_array)
#        self._bunc_sol = solu[8][self._i_solution]
        self._bunc_sol = solu[8]
#        B_var_ap = solu[9][self._i_solution]

        self.set_state(tango.DevState.ON)
        self.set_status('Calculation performed.')
        # PROTECTED REGION END #    //  EEHGMaster.Calculate

    def is_Calculate_allowed(self):
        # PROTECTED REGION ID(EEHGMaster.is_Calculate_allowed) ENABLED START #
        return self.get_state() not in [DevState.RUNNING,DevState.MOVING]
        # PROTECTED REGION END #    //  EEHGMaster.is_Calculate_allowed

    @command(
    )
    @DebugIt()
    def ReadFromField(self):
        # PROTECTED REGION ID(EEHGMaster.ReadFromField) ENABLED START #
        """
        reads from field the attribute `current` values
        :return:Non
        """
        # leggo il primo attributo Lambda fel
        if self.masterfel1_device:
            try:
                lambdaf = self.masterfel1_device.read_attribute("LambdaFEL").value
            except tango.DevFailed:
                lambdaf = 13.3
                # self.error_stream('')
                self.error_stream('failed to read lambda FEL')
                self.set_state(tango.DevState.ALARM)
                self.set_status('failed to read lambda FEL')
            self._lambda_fel_curr = lambdaf
        else:
            self.set_state(tango.DevState.ALARM)
            self.set_status('master fel device not initialized')
            
        # leggo il secondo attributo energy
        if self.sclmagnet_device:
            try:
                valore = self.sclmagnet_device.read_attribute("MagnetEnergy").value
                
            except tango.DevFailed:
                valore = 1.0
                # self.error_stream('')
                self.error_stream('failed to read beam energy')
                self.set_state(tango.DevState.ALARM)
                self.set_status('failed to read beam energy')
                
            self._e_nrg_fel = valore
            # vado a scrivere il limite max di R562 com la formula
            att_pscl = self.r562.get_properties()
            att_pscl.max_value = str((165/valore)**2*0.01469)
            self.r562.set_properties(att_pscl)
            
        else:
            self.set_state(tango.DevState.ALARM)
            self.set_status('scl device not initialized')
        
        # leggo il terzo attributo r561
        if self.delayline_device:
            try:
                if self.delayline_device.read_attribute("State").value != tango.DevState.ON:
                    self.set_state(tango.DevState.ALARM)
                    self.set_status('Delay line is off')
                    valore = 2.0
                else:
                    valore = self.delayline_device.read_attribute("R56").value
                
            except tango.DevFailed:
                valore = 2.0                
                # self.error_stream('')
                self.error_stream('failed to read R561')
                self.set_state(tango.DevState.ALARM)
                self.set_status('Alarm, failed to read R561')
                
            self._r561_curr = valore
            
        else:
            self.set_state(tango.DevState.ALARM)
            self.set_status('delay line device not initialized')
        
        # leggo il terzo attributo r562
        if self.dispmagnet_device:
            try:
                if self.dispmagnet_device.read_attribute("State").value != tango.DevState.ON:
                    self.set_state(tango.DevState.ALARM)
                    self.set_status('Alarm, Dispersive mag. is off')
                    valore = 100.0 
                else:
                    valore = self.dispmagnet_device.read_attribute("R56").value
                
            except tango.DevFailed:
                valore = 100.0                
                # self.error_stream('')
                self.error_stream('failed to read R562')
                self.set_state(tango.DevState.ALARM)
                self.set_status('Alarm, failed to read R562')
                
            self._r562_curr = valore
            
        else:
            self.set_state(tango.DevState.ALARM)
            self.set_status('disp. mag. device not initialized')
        
        # leggo gli attributi del seed
        if self.seedmaster_device:
            try:
                attrs = self.seedmaster_device.read_attributes(self.seed_dev_atts_toread)
                # self.set_state(attrs[0].value)

                if attrs[1].quality != tango.AttrQuality.ATTR_INVALID:
                    self._lambda_s1_curr = attrs[1].value
                else:
                    self._lambda_s1_curr = 260.0
                    
                if attrs[2].quality != tango.AttrQuality.ATTR_INVALID:
                    self._dlambda_s1 = attrs[2].value
                else:
                    self._dlambda_s1 = 1.0
                    
                if attrs[3].quality != tango.AttrQuality.ATTR_INVALID:
                    self._nrg_s1_curr = attrs[3].value
                else:
                    self._nrg_s1_curr = 5.0

                if attrs[4].quality != tango.AttrQuality.ATTR_INVALID:
                    self.nrg_s1_range = attrs[4].value
                    
                    #print('eng. s1 = %f' % self.nrg_s1_range)
                    #self.nrg_s1.set_max_value(self.nrg_s1_range)
                    att_pp1 = self.nrg_s1.get_properties()
                    att_pp1.max_value = str(attrs[4].value)
                    self.nrg_s1.set_properties(att_pp1)
                else:
                    self.nrg_s1_range = 50.0

                if attrs[5].quality != tango.AttrQuality.ATTR_INVALID:
                    self._lambda_s2_curr = attrs[5].value
                else:
                    self._lambda_s2_curr = 255.0

                if attrs[6].quality != tango.AttrQuality.ATTR_INVALID:
                    self._dlambda_s2 = attrs[6].value
                else:
                    self._dlambda_s2 = 1.0

                if attrs[7].quality != tango.AttrQuality.ATTR_INVALID:
                    self._nrg_s2_curr = attrs[7].value
                else:
                    self._nrg_s2_curr = 10.0

                if attrs[8].quality != tango.AttrQuality.ATTR_INVALID:
                    self.nrg_s2_range = attrs[8].value
                    #self.nrg_s2.set_max_value(self.nrg_s2_range)
                    #print('eng. s2 = %f' % self.nrg_s2_range)
                    att_pp = self.nrg_s2.get_properties()
                    att_pp.max_value = str(attrs[8].value)
                    self.nrg_s2.set_properties(att_pp)

                else:
                    self.nrg_s2_range = 40.0

                if attrs[10].quality != tango.AttrQuality.ATTR_INVALID:
#                    il n.10 e` per il seed 2, mentre in n.9 e` per il seed 1.

                    att_pl = self.lambda_s2.get_properties()
                    att_pl.min_value = str(attrs[10].value[0]/2)
                    att_pl.max_value = str(attrs[10].value[1]/2)    # il /due dipende dall'OPA
                    self.lambda_s2.set_properties(att_pl)

            except tango.DevFailed:
                self.set_state(tango.DevState.ALARM)
                self.error_stream('Alarm, seed device failed')
                self.set_status('Alarm, seed device failed')
            except ValueError:
                self.set_state(tango.DevState.ALARM)
                self.error_stream('reading of seed device failed')
                self.set_status('Alarm, reading of seed device failed')

        else:
            self.set_state(tango.DevState.ALARM)
            self.set_status('Alarm, seed device not initialized')

        #print(float(self.lambda_s1.get_properties().min_value) * 1.1)

        # leggo l'attributo waist s1
        if self.seedccd1_device:
            try:
                valore = self.seedccd1_device.read_attribute("HorSigma").value
                
            except tango.DevFailed:
                valore = 0.3              
                # self.error_stream('')
                self.error_stream('failed to read waist 1')
                self.set_state(tango.DevState.ALARM)
                self.set_status('Alarm, failed to read waist 1')
            self._waist_s1 = valore
            
        else:
            self.set_state(tango.DevState.ALARM)
            self.set_status('Alarm, seed ccd1 device not initialized')

        # leggo l'attributo waist s2
        if self.seedccd2_device:
            try:
                valore = self.seedccd2_device.read_attribute("HorSigma").value
                
            except tango.DevFailed:
                valore = 0.2                
                # self.error_stream('')
                self.error_stream('failed to read waist 2')
                self.set_state(tango.DevState.ALARM)
                self.set_status('Alarm, failed to read waist 2')
                
            self._waist_s2 = valore
            
        else:
            self.set_state(tango.DevState.ALARM)
            self.set_status('Alarm, seed ccd2 device not initialized')

        pass
        # PROTECTED REGION END #    //  EEHGMaster.ReadFromField

    def is_ReadFromField_allowed(self):
        # PROTECTED REGION ID(EEHGMaster.is_ReadFromField_allowed) ENABLED START #
        return self.get_state() not in [DevState.RUNNING]
        # PROTECTED REGION END #    //  EEHGMaster.is_ReadFromField_allowed

    @command(
    )
    @DebugIt()
    def ApplyR56s(self):
        # PROTECTED REGION ID(EEHGMaster.ApplyR56s) ENABLED START #
        """
        Applica la soluzione sulle sezioni dispersive
        self._r561, self._r562

        :return:None
        """
        self.set_state(tango.DevState.MOVING)
        self.set_status('In writing R56s')

        try:
#            self.delayline_device.write_attribute("R56", self._r561)
            self.dispmagnet_device.write_attribute("R56", self._r562)
        except tango.DevFailed:
            self.error_stream('Fault: failed to Set R56s')
            self.set_state(tango.DevState.FAULT)
            return -1

        self.set_state(tango.DevState.ON)
        self.set_status('back to ON')
        return 0
        # PROTECTED REGION END #    //  EEHGMaster.ApplyR56s

    def is_ApplyR56s_allowed(self):
        # PROTECTED REGION ID(EEHGMaster.is_ApplyR56s_allowed) ENABLED START #
        return self.get_state() not in [DevState.RUNNING]
        # PROTECTED REGION END #    //  EEHGMaster.is_ApplyR56s_allowed

    @command(
    )
    @DebugIt()
    def ApplySeeds(self):
        # PROTECTED REGION ID(EEHGMaster.ApplySeeds) ENABLED START #
        """
        Applica la soluzione al device master del seed
        from origin:
        targetWL1_nm = self.iperspazio[self.labelApply]['seed1'].wavelength * 1e9
        targetE1_uJ = self.iperspazio[self.labelApply]['seed1'].energy * 1e6
        targetWL2_nm = self.iperspazio[self.labelApply]['seed2'].wavelength * 1e9
        targetE2_uJ = self.iperspazio[self.labelApply]['seed2'].energy * 1e6
        self._lambda_s1 
        self._nrg_s1
        self._lambda_s2 
        self._nrg_s2

        """
        self.set_state(tango.DevState.MOVING)
        self.set_status('In writing Seeds')

        try:
            pass
            self.seedmaster_device.write_attribute("wavelength2", self._lambda_s2)
        except tango.DevFailed:
            self.error_stream('Fault, failed to Set Seeds')
            self.set_state(tango.DevState.FAULT)
            return -1

        self.set_state(tango.DevState.ON)
        self.set_status('back to ON')
        return 0
        # PROTECTED REGION END #    //  EEHGMaster.ApplySeeds

    def is_ApplySeeds_allowed(self):
        # PROTECTED REGION ID(EEHGMaster.is_ApplySeeds_allowed) ENABLED START #
        return self.get_state() not in [DevState.RUNNING]
        # PROTECTED REGION END #    //  EEHGMaster.is_ApplySeeds_allowed

    @command(
    )
    @DebugIt()
    def ApplyIDs(self):
        # PROTECTED REGION ID(EEHGMaster.ApplyIDs) ENABLED START #
        """
        Apply per settare gli ondulatori
        from origin:
        targetFELwl_nm = self.iperspazio['wavelength']['value']
        m = self.iperspazio[self.labelApply]['eehgsetup'].m
        n = self.iperspazio[self.labelApply]['eehgsetup'].n
        virtualHarm = (m + n)
        virtualSeed_nm = targetFELwl_nm * virtualHarm
        # For the modulators
        targetMod1_nm = self.iperspazio[self.labelApply]['seed1'].wavelength * 1e9
        targetMod2_nm = self.iperspazio[self.labelApply]['seed2'].wavelength * 1e9
        :return:None
        to server:
        self._lambda_fel
        self._lambda_s1 /s2
        self.n_value
        self.m_value
        """
        
        virtualHarm = self.m_value + self.n_value
        virtualSeed_nm = self._lambda_fel * virtualHarm
        # For the modulators
        targetMod1_nm = self._lambda_s1 # * 1e9
        targetMod2_nm = self._lambda_s2 # * 1e9

        self.set_state(tango.DevState.MOVING)
        self.set_status('In writing SuperGap')

        try:
            self.masterfel1_device.write_attribute("LambdaSeed", virtualSeed_nm)
            self.masterfel1_device.write_attribute("Harmon", virtualHarm)
            self.mod01fel1_device.write_attribute("LambdaSeed", targetMod1_nm)
            self.mod02fel1_device.write_attribute("LambdaSeed", targetMod2_nm)
        except tango.DevFailed:
            self.error_stream('failed to Set SuperGap')
            self.set_state(tango.DevState.FAULT)
            self.set_status('Fault: failed to set SuperGap')
            return -1
        sleep(1)
        try:
            pass
            self.masterfel1_device.command_inout("ApplyCalculatedValues", 1)
        except tango.DevFailed as e:
            self.error_stream('failed to SuperGap Apply')
            self.set_state(tango.DevState.FAULT)
            self.set_status('Fault: failed to set SuperGap')
            tango.Except.re_throw_exception(e, 'SuperGap Apply failed',
                                            'command_inout failed',
                                            'ApplyCalculatedValues')
            return -1


        self.set_state(tango.DevState.ON)
        self.set_status('ready and ON')
        return 0
    
        # PROTECTED REGION END #    //  EEHGMaster.ApplyIDs

    def is_ApplyIDs_allowed(self):
        # PROTECTED REGION ID(EEHGMaster.is_ApplyIDs_allowed) ENABLED START #
        return self.get_state() not in [DevState.RUNNING]
        # PROTECTED REGION END #    //  EEHGMaster.is_ApplyIDs_allowed

    @command(
    )
    @DebugIt()
    def ApplyAll(self):
        # PROTECTED REGION ID(EEHGMaster.ApplyAll) ENABLED START #
        """
        with this command All Applys will b performed

        :return:None
        """
        
        if self.ApplyR56s():
            self.set_state(tango.DevState.FAULT)
            self.set_status('Fault: failed to set R56')
            return
        
        if self.ApplySeeds():
            self.set_state(tango.DevState.FAULT)
            self.set_status('Fault: failed to set Seeds')
            return
        
        if self.ApplyIDs():
            self.set_state(tango.DevState.FAULT)
            self.set_status('Fault: failed to set IDs')
            
        # PROTECTED REGION END #    //  EEHGMaster.ApplyAll

    def is_ApplyAll_allowed(self):
        # PROTECTED REGION ID(EEHGMaster.is_ApplyAll_allowed) ENABLED START #
        return self.get_state() not in [DevState.RUNNING,DevState.MOVING,DevState.FAULT]
        # PROTECTED REGION END #    //  EEHGMaster.is_ApplyAll_allowed

# ----------
# Run server
# ----------


def main(args=None, **kwargs):
    """Main function of the EEHGMaster module."""
    # PROTECTED REGION ID(EEHGMaster.main) ENABLED START #
    return run((EEHGMaster,), args=args, **kwargs)
    # PROTECTED REGION END #    //  EEHGMaster.main


if __name__ == '__main__':
    main()
